# coding: utf-8
from __future__ import unicode_literals

from jinja2 import Template

from report_writer.core.content.base import BaseContent


try:
    from typing import TYPE_CHECKING
except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
    from typing import Any
    from typing import Type
    from typing import Tuple


class ContextContent(BaseContent):
    """
    Контент описывающий прямое использование контекста.

    Не производит преобразований.

    .. code-block:: python

        content = ContextContent()
        content.render('foo')  # foo
        content.render(1)  # 1
        content.render(None)  # None

    """
    def render(self,
               context  # type: Any
               ):  # type: (...) -> Any
        return context


class KeyContent(BaseContent):
    """
    Контент описывающий получение данных по ключу
    из словаря и подобных объектов.

    Можно передавать более 1 ключа для получения данных из вложенных словарей.

    .. code-block:: python

        content = KeyContent('foo')
        content.render({'foo': 1})  # 1

        content = KeyContent('foo', 'bar')
        content.render({'foo': {'bar': 1}})  # 1

        content = KeyContent('foo', default=1)
        content.render({'bar': 2})  # 1

    :param keys: Ключ(и) словаря
    :param opt: Дополнительные параметры
    """
    def __init__(self,
                 *keys,  # type: str
                 **opt  # type: Any
                 ):  # type: (...) -> None
        self._default = opt.get('default')  # type: Any
        self._has_default = 'default' in opt  # type: bool
        self._keys = keys  # type: Tuple[str, ...]

    def render(self,
               context  # type: Any
               ):  # type: (...) -> Any
        current = context

        try:
            for key in self._keys:
                current = current[key]
        except KeyError:
            if self._has_default:
                return self._default
            raise KeyError(', '.join(self._keys))

        return current


class AttrContent(BaseContent):
    """
    Контент описывающий получение данных по названию атрибута.

    Можно передавать более 1 названия атрибута для получения данных из
    вложенных объектов.

    .. code-block:: python

        content = AttrContent('foo')
        content.render(namedtuple('Object', 'foo')(1))  # 1

        content = AttrContent('foo', 'bar')
        content.render(
            namedtuple('Object', 'foo')(
                namedtuple('Object', 'bar')(1)
            )
        )  # 1

        content = AttrContent('foo', default=1)
        content.render(namedtuple('Object', 'bar')(2))  # 1

    :param keys: Атрибут(ы) объекта
    :param opt: Дополнительные параметры
    """
    def __init__(self,
                 *keys,  # type: str
                 **opt  # type: Any
                 ):  # type: (...) -> None
        self._default = opt.get('default')  # type: Any
        self._has_default = 'default' in opt  # type: bool
        self._keys = keys  # type: Tuple[str, ...]

    def render(self,
               context  # type: Any
               ):  # type: (...) -> Any
        current = context

        try:
            for key in self._keys:
                current = getattr(current, key)
        except AttributeError:
            if self._has_default:
                return self._default
            raise AttributeError(', '.join(self._keys))

        return current


class IndexContent(BaseContent):
    """
    Контент описывающий получение данных по индексу
    из массива и подобных объектов.

    .. code-block:: python

        content = IndexContent(1)
        content.render([0, 1])  # 1

        content = IndexContent(1, default=1)
        content.render([])  # 1

    :param index: Индекс для получения данных из коллекции
    :param opt: Дополнительные параметры
    """
    def __init__(self,
                 index,  # type: int
                 **opt  # type: Any
                 ):  # type: (...) -> None
        self._default = opt.get('default')  # type: Any
        self._has_default = 'default' in opt  # type: bool
        self._index = index  # type: int

    def render(self,
               context  # type: Any
               ):  # type: (...) -> Any
        try:
            return context[self._index]
        except IndexError:
            if self._has_default:
                return self._default
            raise


class TemplateContent(BaseContent):
    """
    Контент описывающий получение данных по шаблону
    из словаря и подобных объектов.

    Для рендера шаблона используется jinja2.

    .. code-block:: python

        content = TemplateContent('{{ foo }}')
        content.render({'foo': 1})  # '1'
        content.render({'bar': 1})  # ''

        content = TemplateContent(
            '{% if foo %}{{ foo }}{% else %}No{% endif %}'
        )
        content.render({'foo': 1})  # '1'
        content.render({'bar': 1})  # 'No'

    :param template_str: Строка шаблона
    """
    template_cls = Template  # type: Type[Template]

    def __init__(self,
                 template_str  # type: str
                 ):  # type: (...) -> None
        self._template = self.get_template(template_str)  # type: Template

    def get_template(self,
                     template_str  # type: str
                     ):  # type: (...) -> Template
        """
        Получение шаблона.

        :param template_str: Строка шаблона
        :return: Шаблон
        """
        return self.template_cls(template_str)

    @staticmethod
    def render_template(template,  # type: Template
                        context  # type: dict
                        ):  # type: (...) -> str
        """
        Рендер шаблона.

        :param template: Шаблон
        :param context: Контекст рендера шаблона
        :return: Отрендеренная строка
        """
        return template.render(context)

    def render(self,
               context  # type: Any
               ):  # type: (...) -> Any
        return self.render_template(self._template, context)
