# coding: utf-8
from __future__ import unicode_literals

from itertools import groupby
from itertools import islice

from six.moves import zip_longest

from report_writer.core.content.base import BaseContent
from report_writer.core.content.samples.getters import ContextContent
from report_writer.utils.iterators import with_next


try:
    from typing import TYPE_CHECKING
except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
    from typing import Any
    from typing import Callable
    from typing import Iterable
    from typing import Optional


class ContentIterator(BaseContent):
    """
    Контент описывающий получение итератора данных из другого итератора.

    Каждый элемент переданного контекста-итератора преобразуется с
    использованием переданного базового контента

    .. code-block:: python

        content = ContentIterator(KeyContent('foo'))
        content.render([
            {'foo': 1},
            {'foo': 3, 'bar': 5},
            {'foo': 5},
            {'foo': 7},
        ])  # [1, 3, 5, 7]

    :param base_content: Базовый контент для элементов последовательности
    """

    default_base_content = ContextContent()  # type: BaseContent

    def __init__(self,
                 base_content=None  # type: BaseContent
                 ):  # type: (...) -> None
        if base_content is None:
            base_content = self.default_base_content

        self._base_content = base_content  # type: BaseContent

    def render(self,
               context  # type: Any
               ):  # type: (...) -> Any
        for _context in context:
            yield self._base_content.render(_context)


class ContentBlock(BaseContent):
    """
    Контент описывающий получение итератора данных с использованием
    итератора контента.

    Каждый элемент итератора объектов контента получает данные из
    переданного контекста.

    .. code-block:: python

        content = ContentBlock([
            KeyContent('foo'),
            KeyContent('bar'),
            KeyContent('baz'),
        ])
        content.render({
            'foo': 1,
            'bar': 3,
            'baz': 5,
        })  # [1, 3, 5]

    :param content_iterator: Итератор объектов контента для блока
    """
    def __init__(self,
                 content_iterator  # type: Iterable[BaseContent]
                 ):  # type: (...) -> None
        self._content_iterator = (
            content_iterator)  # type: Iterable[BaseContent]

    def render(self,
               context  # type: Any
               ):  # type: (...) -> Any
        for content in self._content_iterator:
            yield content.render(context)


class SizedContentIterator(ContentIterator):
    """
    То же самое что и :code:`ContentIterator`, но ограниченного размера.

    .. code-block:: python

        content = SizedContentIterator(size=5)
        # [0, 1, 2, 3, 4]
        content.render(range(100))

    :param base_content: Базовый контент для элементов последовательности
    :param size: Используемое количество элементов итератора
    """
    def __init__(self,
                 base_content=None,  # type: BaseContent
                 size=None  # type: Optional[int]
                 ):  # type: (...) -> None
        super(SizedContentIterator, self).__init__(base_content)
        self._size = size  # type: Optional[int]

    def render(self,
               context  # type: Any
               ):  # type: (...) -> Any
        for result in super(SizedContentIterator, self).render(
                islice(context, self._size)):
            yield result


class BorderedContentIterator(ContentIterator):
    """
    То же самое что и :code:`ContentIterator`, но для каждого элемента
    возвращает так же данные о том является ли он первым и/или последним

    .. code-block:: python

        content = BorderedContentIterator()
        # [
        #     (0, True, False),
        #     (1, False, False),
        #     (2, False, False),
        #     (3, False, False),
        #     (4, False, True),
        # ]
        content.render(range(5))
        content.render([1])  # [(1, True, True)]

    """
    def render(self,
               context  # type: Any
               ):  # type: (...) -> Any
        first = True
        none = object()
        for result, _next in with_next(
            super(BorderedContentIterator, self).render(context),
            none
        ):
            yield result, first, _next is none
            if first:
                first = False


class GroupedContentIterator(ContentIterator):
    """
    Контент описывающий получение сгруппированных данных из итератора.

    Группирует данные с помощью переданной функции группировки, и возвращает
    кортежи из ключа группировки и группы.

    .. code-block:: python

        content = GroupedContentIterator(group_by=itemgetter(0))

        # [
        #     (0, ((0, 0), (0, 1))),
        #     (1, ((1, 0), (1, 1), (1, 2)))
        # ]
        content.render([
            (0, 0),
            (0, 1),
            (1, 0),
            (1, 1),
            (1, 2),
        ])

    :param base_content: Базовый контент для элементов последовательности
    :param group_by: Функция группировки
    """
    def __init__(self,
                 base_content=None,  # type: BaseContent
                 key_content=None,  # type: BaseContent
                 group_by=None  # type: Optional[Callable[[Any], Any]]
                 ):  # type: (...) -> None
        super(GroupedContentIterator, self).__init__(base_content)
        if key_content is None:
            key_content = self.default_base_content

        self._group_by = group_by  # type: Optional[Callable[[Any], Any]]
        self._key_content = key_content  # type: BaseContent

    def render(self,
               context  # type: Any
               ):  # type: (...) -> Any
        for key, group_iterator in groupby(context, self._group_by):
            yield (
                self._key_content.render(key),
                self._base_content.render(group_iterator)
            )


class BorderedGroupedContentIterator(GroupedContentIterator):
    """
    То же самое что и :code:`GroupedContentIterator`, но для каждой
    преобразованной группы возвращает так же данные о том является ли она
    первой и/или последней.

    .. code-block:: python

        content = BorderedGroupedContentIterator(group_by=itemgetter(0))

        # [
        #     (0, (
        #         ((0, 0), True, False),
        #         ((0, 1), False, True)
        #     )),
        #     (1, (
        #         ((1, 0), True, False),
        #         ((1, 1), False, False),
        #         ((1, 2), False, True)
        #     ))
        #     (2, (
        #         ((2, 0), True, True)
        #     )),
        # ]
        content.render([
            (0, 0),
            (0, 1),
            (1, 0),
            (1, 1),
            (1, 2),
            (2, 0),
        ])
    """
    def render(self,
               context  # type: Any
               ):  # type: (...) -> Any
        first = True
        none = object()

        for (key, group_iterator), _next in with_next((
            (k, tuple(g))
            for k, g in groupby(context, self._group_by)),
            none
        ):
            yield (
                key,
                self._base_content.render(group_iterator)
            ), first, _next is none

            if first:
                first = False


class ContentZip(ContentIterator):
    """
    Контент описывающий получение данных с помощью builtin функции zip.

    .. code-block:: python

        content = ContentZip(
            None,
            [
                KeyContent('foo'),
                KeyContent('bar'),
                KeyContent('baz'),
            ]
        )
        # [
        #     (1, 4, 6),
        #     (2, 5, 7),
        # ]
        content.render({
            'foo': [1, 2, 3],
            'bar': [4, 5],
            'baz': [6, 7, 8, 9],
        })

        content = ContentZip(
            None,
            [
                KeyContent('foo'),
                KeyContent('bar'),
                KeyContent('baz'),
            ],
            longest=True
        )
        # [
        #     (1, 4, 6),
        #     (2, 5, 7),
        #     (3, None, 8),
        #     (None, None, 9),
        # ]
        content.render({
            'foo': [1, 2, 3],
            'bar': [4, 5],
            'baz': [6, 7, 8, 9],
        })

    :param base_content: Базовый контент для элементов последовательности
    :param iterators: Объекты итераторов данных для получения значений
                      передаваемых в zip
    :param longest: Использовать zip_longest
    """
    def __init__(self,
                 base_content,  # type: BaseContent
                 iterators,  # type: Iterable[BaseContent]
                 longest=False  # type: bool
                 ):  # type: (...) -> None
        super(ContentZip, self).__init__(base_content)
        self._iterators = iterators  # type: Iterable[BaseContent]
        self._longest = longest  # type: bool

    def render(self,
               context  # type: Any
               ):  # type: (...) -> Any
        iterators = (
            content.render(context)
            for content in self._iterators
        )

        _zip = zip
        if self._longest:
            _zip = zip_longest

        for group in _zip(*iterators):
            yield self._base_content.render(group)


class ContentMapping(ContentBlock):
    """
    Контент описывающий получение итератора данных с помощью мэппинга объектов
    контента из переданного итератора объектов контента и элементов переданного
    контекста-итератора.

    .. code-block:: python

        content = ContentMapping([
            KeyContent('foo'),
            KeyContent('bar'),
            KeyContent('baz'),
        ])
        content.render([
            {'foo': 0, 'bar': 1, 'baz': 2},
            {'foo': 2, 'bar': 0, 'baz': 1},
            {'foo': 1, 'bar': 2, 'baz': 0},
        ])  # [0, 0, 0]

    """
    def render(self,
               context  # type: Any
               ):  # type: (...) -> Any
        for content, _context in zip(self._content_iterator, context):
            yield content.render(_context)


class ContentFilter(ContentIterator):
    """
    Контент описывающий получение данных с помощью builtin функции filter.

    .. code-block:: python

        content = ContentFilter(
            KeyContent('foo'),
            lambda el: 'foo' in el
        )
        # [0, 1, 2]
        content.render([
            {'foo': 0},
            {'bar': 0},
            {'foo': 1, 'bar': 1},
            {'foo': 2},
            {'bar': 2},
        ])

    :param base_content: Базовый контент для элементов последовательности
    :param filter_fn: Функция фильтрации
    """
    def __init__(self,
                 base_content=None,  # type: BaseContent
                 filter_fn=None  # type: Optional[Callable[[Any], bool]]
                 ):  # type: (...) -> None
        super(ContentFilter, self).__init__(base_content)
        self._filter_fn = filter_fn  # type: Optional[Callable[[Any], bool]]

    def render(self,
               context  # type: Any
               ):  # type: (...) -> Any
        for content in filter(self._filter_fn,
                              super(ContentFilter, self).render(context)):
            yield content


class EnumerateContent(ContentIterator):
    """
    Контент описывающий получение данных с помощью builtin функции enumerate.

    .. code-block:: python

        content = EnumerateContent(KeyContent('foo'), start=1)
        content.render([
            {'foo': 0},
            {'foo': 2},
            {'foo': 4},
        ])  # [(1, 0), (2, 2), (3, 4)]

    :param base_content: Базовый контент для элементов последовательности
    :param start: Стартовый порядковый номер
    """

    def __init__(self,
                 base_content=None,  # type: BaseContent
                 start=0,  # type: int
                 ):  # type: (...) -> None
        super(EnumerateContent, self).__init__(base_content)
        self._start = start  # type: int

    def render(self,
               context  # type: Any
               ):  # type: (...) -> Any
        for content in enumerate(
                super(EnumerateContent, self).render(context),
                start=self._start):
            yield content
