# coding: utf-8
from __future__ import unicode_literals

import re
from collections import defaultdict

import six
from jinja2 import Template
from openpyxl import Workbook
from openpyxl import load_workbook

from report_writer.core.report_objects.cursor import Cursor


try:
    from typing import TYPE_CHECKING
except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
    from typing import Any
    from typing import DefaultDict
    from typing import Dict
    from typing import Optional
    from typing import Tuple
    from typing import Type

    from openpyxl.cell import Cell
    from openpyxl.styles import NamedStyle
    from openpyxl.worksheet.worksheet import Worksheet


class TemplateFile(object):
    """
    Шаблон отчёта.

    :param path: Путь до файла аблона
    :param context: Контекст рендера шаблона
    """

    # Дефолтные курсоры для шаблона можно указывать (внутри клетки) тэгом
    # cursor#[<cursor_name>], а потом получать с помощью <cursor_name>
    # cursor_name - валидный идентификатор
    # Например: cursor#[main]
    cursor_tag = re.compile(
        r'cursor#\[(?P<cursor_name>[_a-zA-Z][_a-zA-Z0-9]*)\]')

    def __init__(self,
                 path,  # type: str
                 context  # type: Dict[str, Any]
                 ):  # type: (...) -> None
        self._path = path
        self._context = context
        self._workbook = None  # type: Optional[Workbook]

        self._all_cursors = defaultdict(
            dict
        )  # type: DefaultDict[Optional[str], Dict[str, Tuple[int, int]]]

        self._init_workbook()

    def get_cursor_name(self,
                        cell  # type: Cell
                        ):  # type: (...) -> Optional[str]
        """
        Получение названия курсора из клетки если доступно

        :param cell: Клетка файла шаблона
        :return: Если есть - название курсора помеченного в шаблоне
        """
        if isinstance(cell.value, six.string_types):
            text = cell.value
            match = self.cursor_tag.match(text)
            if match is not None:
                cursor_name = match.group('cursor_name')
                return cursor_name

        return None

    def render_cell(self,
                    cell  # type: Cell
                    ):  # type: (...) -> None
        """
        Рендер содержимого клетки если доступно

        :param cell: Клетка файла шаблона
        """
        if isinstance(cell.value, six.string_types):
            cell.value = Template(cell.value).render(
                **self._context)

    def _init_workbook(self
                       ):  # type: () -> None
        """
        Инициализация объекта Workbook.
        """
        wb = load_workbook(self._path)  # type: Workbook

        for ws in wb:
            for row_idx, row in enumerate(ws.__iter__()):
                for col_idx, cell in enumerate(row):
                    self.render_cell(cell)

                    cursor_name = self.get_cursor_name(cell)
                    if cursor_name is not None:
                        self._all_cursors[ws.title][cursor_name] = (
                            row_idx, col_idx)
                        cell.value = None

        self._workbook = wb

    @property
    def workbook(self
                 ):  # type: () -> Workbook
        """
        :return: Workbook из шаблона
        """
        return self._workbook

    def get_cursor(self,
                   ws_name,  # type: str
                   cursor_name  # type: str
                   ):  # type: (...) -> Cursor
        """
        Получение курсора из шаблона по имени.

        :param ws_name: Название листа
        :param cursor_name: Название курсора
        :return: Именованный курсор из данного листа
        """
        return Cursor(*self._all_cursors[ws_name][cursor_name])


class ReportFile(object):
    """
    Файл отчёта для работы с openpyxl

    :param path: Путь до файла отчёта
    :param template: Путь до шаблона отчёта
    :param context: Контекст рендера шаблона
    """

    template_file_cls = TemplateFile  # type: Type[TemplateFile]

    def __init__(self,
                 path,  # type: Optional[str]
                 template=None,  # type: Optional[str]
                 context=None  # type: Optional[dict]
                 ):  # type: (...) -> None
        self._path = path
        self._template_file = None  # type: Optional[TemplateFile]

        if template is not None:
            self._template_file = self._get_template_file(
                template, context or {})

        self._workbook = None  # type: Optional[Workbook]
        self._worksheets = {}  # type: Dict[Optional[str], Worksheet]

        self.init_workbook()

    def _get_template_file(self,
                           template,  # type: str
                           context,  # type: Dict[str, Any]
                           ):  # type: (...) -> TemplateFile
        """
        Получение шаблона отчёта.

        :param template: Путь до шаблона отчёта
        :param context: Контекст рендера шаблона отчёта
        :return: Файл шаблона отчёта
        """
        return self.template_file_cls(template, context)

    def _get_workbook(self
                      ):  # type: () -> Workbook
        """
        Получение объекта Workbook.

        :return: Объект Workbook из шаблона или созданный с нуля
        """
        if self._template_file is not None:
            wb = self._template_file.workbook
        else:
            wb = Workbook()

        return wb

    def init_workbook(self
                      ):  # type: () -> None
        """
        Инициализация объекта Workbook для файла.
        """
        self._workbook = self._get_workbook()
        active = self.workbook.active

        for worksheet in self.workbook:
            # Вносим все существующие листы, используя в качестве
            # ключа их названия
            self._worksheets[worksheet.title] = worksheet

        self.set_worksheet(active.title)

    @property
    def workbook(self
                 ):  # type: () -> Workbook
        """
        :return: Объект Workbook файла
        """
        return self._workbook

    @property
    def worksheet(self
                  ):  # type: () -> Worksheet
        """
        :return: Текущий активный лист файла
        """
        return self.workbook.active

    def add_named_style(self,
                        style  # type: NamedStyle
                        ):  # type: (...) -> None
        """
        Добавление именованного стиля к файлу.

        :param style: Именованный стиль
        """
        self.workbook.add_named_style(style)

    def set_worksheet(self,
                      name=None,  # type: Optional[str]
                      title=None,  # type: Optional[str]
                      idx=None  # type: Optional[int]
                      ):  # type: (...) -> Worksheet
        """
        Смена листа либо создание нового.

        Для получения листа из шаблона стоит обращаться по заголовкам листков

        :param name: Название листа
        :param title: Заголовок листа
        :param idx: Индекс расположения листа
        :return: Существующий лист или созданный
        """
        if name in self._worksheets:
            self.workbook.active = self._worksheets[name]
        else:
            self._worksheets[name] = self.workbook.create_sheet(title, idx)

        return self._worksheets[name]

    def get_cursor(self,
                   name  # type: str
                   ):  # type: (...) -> Cursor
        """
        Получение курсора из шаблона по имени.

        :param name: Название курсора
        :return: Именованный курсор по названию
        """
        # Т.к. именнованные курсоры могут прийти только из шаблона,
        # обращение будет по названию листа, а не по ключу словаря

        if self._template_file is None:
            raise ValueError('Нельзя получить именованные курсоры если '
                             'файл отчёта не использует шаблонов')

        return self._template_file.get_cursor(
            self.workbook.active.title, name)

    def save(self
             ):  # type: () -> None
        self.workbook.save(self._path)
        self.workbook.close()
