# coding: utf-8
from __future__ import division
from __future__ import unicode_literals

import abc

import six
from openpyxl.utils import get_column_letter

from report_writer.impl.xlsx.report.writers.base import BaseReportWriter
from report_writer.impl.xlsx.report.writers.merge import MergedCellWriter
from report_writer.impl.xlsx.report.writers.simple import CellWriter
from report_writer.impl.xlsx.utils.coordinates import get_coordinates


try:
    from typing import TYPE_CHECKING
except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
    from typing import Any
    from typing import Dict
    from typing import Iterable
    from typing import Union

    from openpyxl.styles import NamedStyle
    from openpyxl.styles import Font
    from openpyxl.styles import Fill
    from openpyxl.styles import Border
    from openpyxl.styles import Alignment
    from openpyxl.styles import Protection
    from openpyxl.styles.styleable import StyleableObject

    from report_writer.impl.xlsx.report.report_file import ReportFile

    Style = Union[NamedStyle, Font, Fill, Border, Alignment, Protection, str]


class NamedStyleWriter(BaseReportWriter):
    """
    Писатель для именованного стиля.

    :param report: Файл отчёта
    :param style: Именованный стиль для записи
    """
    def __init__(self,
                 report,  # type: ReportFile
                 style  # type: NamedStyle
                 ):  # type: (...) -> None
        super(NamedStyleWriter, self).__init__(report)
        self._style = style  # type: NamedStyle

    def write(self,
              row=0,  # type: int
              col=0,  # type: int
              data=None  # type: Any
              ):  # type: (...) -> None
        self._report.add_named_style(self._style)


class StyledWriterMixin(object):
    """
    Базовый писатель для использования стилей.
    """

    style_attrs = (
        'font',
        'fill',
        'border',
        'alignment',
        'number_format',
        'protection',
        'style',
    )

    @classmethod
    def apply_styles(cls,
                     cells,  # type: Iterable[StyleableObject]
                     styles  # type: Dict[str, Style]
                     ):
        # type: (...) -> None
        """
        Применение стилей к переданным клеткам
        """
        for cell in cells:
            for style_attr in cls.style_attrs:
                style = styles.get(style_attr)
                if style is not None:
                    setattr(cell, style_attr, style)


class StyledCellWriter(StyledWriterMixin, CellWriter):
    """
    Писатель стилизованной клетки.

    :param report: Файл отчёта
    :param styles: Параметры стиля клетки
    """
    def __init__(self,
                 report,  # type: ReportFile
                 **styles  # type: Style
                 ):  # type: (...) -> None
        super(StyledCellWriter, self).__init__(report)
        self._styles = styles  # type: Dict[str, Style]

    def write(self,
              row,  # type: int
              col,  # type: int
              data  # type: Any
              ):  # type: (...) -> None
        cell = self._report.worksheet[get_coordinates(row, col)]
        self.apply_styles((cell,), self._styles)
        super(StyledCellWriter, self).write(row, col, data)


class StyledMergedCellWriter(StyledWriterMixin, MergedCellWriter):
    """
    Писатель стилизованной объединённой клетки.

    :param report: Файл отчёта
    :param width: Ширина объединения
    :param height: Высота объединения
    :param styles: Параметры стиля клетки
    """
    def __init__(self,
                 report,  # type: ReportFile
                 height,  # type: int
                 width,  # type: int
                 **styles  # type: Style
                 ):  # type: (...) -> None
        super(StyledMergedCellWriter, self).__init__(report, height, width)
        self._styles = styles  # type: Dict[str, Style]

    def write(self,
              row,  # type: int
              col,  # type: int
              data  # type: Any
              ):  # type: (...) -> None
        # Применяем стили и к верхней и к нижней клетке, т.к. если применить
        # только к верхней не будут учитываться нижние границы
        to_row, to_col = row + self._height - 1, col + self._width - 1
        top_cell = self._report.worksheet[get_coordinates(row, col)]
        bottom_cell = self._report.worksheet[get_coordinates(to_row, to_col)]

        self.apply_styles((top_cell, bottom_cell), self._styles)

        super(StyledMergedCellWriter, self).write(row, col, data)


@six.add_metaclass(abc.ABCMeta)
class ColumnBaseWriter(BaseReportWriter):
    """
    Писатель колонки.

    :param report: Файл отчёта
    :param cols: Сколько колонок охватывает настройка
    """

    def __init__(self,
                 report,  # type: ReportFile
                 cols=1,  # type: int
                 ):  # type: (...) -> None
        super(ColumnBaseWriter, self).__init__(report)
        self._cols = cols  # type: int


class ColumnWidthWriter(ColumnBaseWriter):
    """
    Писатель ширины колонки.
    """
    def write(self,
              row,  # type: int
              col,  # type: int
              data  # type: Any
              ):  # type: (...) -> None
        self._report.worksheet.column_dimensions[
            get_column_letter(col + 1)].width = data
        self._report.worksheet.column_dimensions[
            get_column_letter(col + 1)].max = col + self._cols


class ColumnStyleWriter(StyledWriterMixin, ColumnBaseWriter):
    """
    Писатель стиля колонки

    :param report: Файл отчёта
    :param cols: Сколько колонок охватывает настройка
    :param styles: Стили колонки
    """

    def __init__(self,
                 report,  # type: ReportFile
                 cols=1,  # type: int
                 **styles  # type: Style
                 ):  # type: (...) -> None
        super(ColumnStyleWriter, self).__init__(report, cols)
        self._styles = styles  # type: Dict[str, Style]

    def write(self,
              row,  # type: int
              col,  # type: int
              data=None  # type: Any
              ):  # type: (...) -> None
        column_dimension = self._report.worksheet.column_dimensions[
            get_column_letter(col + 1)]

        self.apply_styles([column_dimension], self._styles)

        self._report.worksheet.column_dimensions[
            get_column_letter(col + 1)].max = col + self._cols


class RowHeightWriter(BaseReportWriter):
    """
    Писатель высоты строки.
    """
    def write(self,
              row,  # type: int
              col,  # type: int
              data  # type: Any
              ):  # type: (...) -> None
        self._report.worksheet.row_dimensions[row + 1].height = data


class RowStyleWriter(StyledWriterMixin, BaseReportWriter):
    """
    Писатель стиля строки.

    :param report: Файл отчёта
    :param styles: Стили строки
    """

    def __init__(self,
                 report,  # type: ReportFile
                 **styles  # type: Style
                 ):  # type: (...) -> None
        super(RowStyleWriter, self).__init__(report)
        self._styles = styles  # type: Dict[str, Style]

    def write(self,
              row,  # type: int
              col,  # type: int
              data=None  # type: Any
              ):  # type: (...) -> None
        row_dimension = self._report.worksheet.row_dimensions[row + 1]

        self.apply_styles([row_dimension], self._styles)
