# coding: utf-8
from __future__ import unicode_literals

import abc

import six

from report_writer.core.writers.base import BaseWriter


try:
    from typing import TYPE_CHECKING
except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
    from typing import Any
    from typing import Dict
    from typing import List
    from typing import Set
    from typing import Tuple


class ReportFileMock(object):
    """
    "Файл" отчёта.

    Вместо .xlsx файла производит запись в двумерный массив
    """
    def __init__(self,
                 h,  # type: int
                 w  # type: int
                 ):  # type: (...) -> None
        self.w = w
        self.h = h

        self.data = [
            [
                None
                for _ in range(w)
            ]
            for _ in range(h)
        ]  # type: List[List[Any]]

        self.merged = []  # type: List[Set[Tuple[int, int]]]
        self.formats = {}  # type: Dict[str, Any]

        self._rows = {}  # type: Dict[int, float]
        self._cols = {}  # type: Dict[int, float]

        self.saved = False

    def add_format(self,
                   name,  # type: str
                   format_object  # type: Any
                   ):  # type: (...) -> None
        """Добавление именованного формата."""
        self.formats[name] = format_object

    def get_format(self,
                   name  # type: str
                   ):  # type: (...) -> Any
        """Получение именованного формата."""
        return self.formats.get(name)

    def write(self,
              row,  # type: int
              col,  # type: int
              data  # type: Any
              ):  # type: (...) -> None
        """Запись в клетку."""
        self.data[row][col] = data

    def merge(self,
              from_row,  # type: int
              from_col,  # type: int
              to_row,  # type: int
              to_col  # type: int
              ):  # type: (...) -> None
        """Добавление объединения клеток."""
        r = set()
        for row in range(from_row, to_row + 1):
            for col in range(from_col, to_col + 1):
                r.add((row, col))

        self.merged.append(r)

    def set_format(self,
                   row,  # type: int
                   col,  # type: int
                   format_  # type: Any
                   ):  # type: (...) -> None
        """Смена формата клетки."""
        v = self.data[row][col]
        self.data[row][col] = (format_, v)


@six.add_metaclass(abc.ABCMeta)
class ReportWriter(BaseWriter):
    """
    Писатель отчёта.
    """
    def __init__(self,
                 report  # type: ReportFileMock
                 ):  # type: (...) -> None
        self.report = report


class CellWriter(ReportWriter):
    """
    Писатель клетки отчёта.
    """
    def write(self,
              row,  # type: int
              col,  # type: int
              data  # type: Any
              ):  # type: (...) -> None
        self.report.write(row, col, data)


class MergeWriter(ReportWriter):
    """
    Писатель объединения.
    """
    def __init__(self,
                 report,  # type: ReportFileMock
                 height,  # type: int
                 width  # type: int
                 ):  # type: (...) -> None
        super(MergeWriter, self).__init__(report)
        self.height = height
        self.width = width

    def write(self,
              row,  # type: int
              col,  # type: int
              data  # type: Any
              ):  # type: (...) -> None
        self.report.merge(
            row, col, row + self.height - 1, col + self.width - 1)


class MergedCellWriter(CellWriter):
    """
    Писатель объединённой клетки отчёта.
    """
    def __init__(self,
                 report,  # type: ReportFileMock
                 height,  # type: int
                 width  # type: int
                 ):  # type: (...) -> None
        super(MergedCellWriter, self).__init__(report)
        self.merge_writer = MergeWriter(self.report, height, width)

    def write(self,
              row,  # type: int
              col,  # type: int
              data  # type: Any
              ):  # type: (...) -> None
        self.merge_writer.write(row, col, None)
        super(MergedCellWriter, self).write(row, col, data)


class NamedFormatWriter(BaseWriter):
    """
    Писатель именованного формата.
    """
    def __init__(self,
                 report,  # type: ReportFileMock
                 format_name,  # type: str
                 format_object  # type: Any
                 ):  # type: (...) -> None
        self.report = report
        self.format_name = format_name
        self.format_object = format_object

    def write(self,
              row=0,  # type: int
              col=0,  # type: int
              data=None  # type: Any
              ):  # type: (...) -> None
        self.report.add_format(self.format_name, self.format_object)


class FormattedCellWriter(CellWriter):
    """
    Писатель отформатированной клетки отчёта.
    """
    def __init__(self,
                 report,  # type: ReportFileMock
                 format_name  # type: str
                 ):  # type: (...) -> None
        super(FormattedCellWriter, self).__init__(report)
        self.format_name = format_name

    def write(self,
              row,  # type: int
              col,  # type: int
              data  # type: Any
              ):  # type: (...) -> None
        super(FormattedCellWriter, self).write(row, col, data)
        self.report.set_format(row, col,
                               self.report.get_format(self.format_name))


class FormattedMergedCellWriter(MergedCellWriter):
    """
    Писатель отформатированная объединённой клетки отчёта.
    """
    def __init__(self,
                 report,  # type: ReportFileMock
                 height,  # type: int
                 width,  # type: int
                 format_name  # type: str
                 ):  # type: (...) -> None
        super(FormattedMergedCellWriter, self).__init__(report, height, width)
        self.format_name = format_name

    def write(self,
              row,  # type: int
              col,  # type: int
              data  # type: Any
              ):  # type: (...) -> None
        super(FormattedMergedCellWriter, self).write(row, col, data)
        self.report.set_format(row, col,
                               self.report.get_format(self.format_name))
